@@ -75,9 +75,7 @@ static enum OperandShape operand_shape(const struct HooksAndData *hd)
 {
   U8 lhs_gimme;
   switch(hd->hooks->lhs_flags & 0x07) {
-    case 0: /* back-compat */
-    case XPI_OPERAND_ARITH:
-    case XPI_OPERAND_TERM:
+    case 0:
       lhs_gimme = G_SCALAR;
       break;
 
@@ -93,9 +91,7 @@ static enum OperandShape operand_shape(const struct HooksAndData *hd)
 
   U8 rhs_gimme;
   switch(hd->hooks->rhs_flags & 0x07) {
-    case 0: /* back-compat */
-    case XPI_OPERAND_ARITH:
-    case XPI_OPERAND_TERM:
+    case 0:
       rhs_gimme = G_SCALAR;
       break;
 
@@ -150,10 +146,16 @@ struct Registration {
 
 static struct Registration *registrations;
 
-static OP *new_op(pTHX_ const struct HooksAndData hd, U32 flags, OP *lhs, OP *rhs)
+static OP *new_op(pTHX_ const struct HooksAndData hd, U32 flags, OP *lhs, OP *rhs, ANY *parsedata)
 {
-  if(hd.hooks->new_op)
-    return (*hd.hooks->new_op)(aTHX_ flags, lhs, rhs, hd.data);
+  if(hd.hooks->new_op) {
+    if(hd.hooks->flags & (1<<15)) {
+      OP *(*new_op_v1)(pTHX_ U32, OP *, OP *, void *) = (OP *(*)(pTHX_ U32, OP *, OP *, void *))hd.hooks->new_op;
+      return (*new_op_v1)(aTHX_ flags, lhs, rhs, hd.data); /* no parsedata */
+    }
+
+    return (*hd.hooks->new_op)(aTHX_ flags, lhs, rhs, parsedata, hd.data);
+  }
 
   OP *ret = newBINOP_CUSTOM(hd.hooks->ppaddr, flags, lhs, rhs);
 
@@ -231,51 +233,40 @@ static OP *S_unwrap_list(pTHX_ OP *o, bool may_unwrap_anonlist)
 
 #ifdef HAVE_PL_INFIX_PLUGIN
 
-OP *parse(pTHX_ OP *lhs, struct Perl_custom_infix *def)
+void parse(pTHX_ ANY *parsedata, struct Perl_custom_infix *def)
+{
+  struct Registration *reg = (struct Registration *)def;
+
+  (*reg->hd.hooks->parse)(aTHX_ 0, parsedata, reg->hd.data);
+}
+
+OP *build_op(pTHX_ ANY *parsedata, OP *lhs, OP *rhs, struct Perl_custom_infix *def)
 {
   struct Registration *reg = (struct Registration *)def;
 
   switch(reg->hd.hooks->lhs_flags & 0x07) {
-    case XPI_OPERAND_TERM:
+    case 0:
       break;
 
     case XPI_OPERAND_TERM_LIST:
+    case XPI_OPERAND_LIST:
       lhs = force_list_keeping_pushmark(lhs);
       break;
   }
 
   /* TODO: maybe operator has a 'parse' hook? */
 
-  lex_read_space(0);
-  OP *rhs = NULL;
-
-  switch(reg->hd.hooks->rhs_flags & 0x87) {
-    case XPI_OPERAND_ARITH:
-      rhs = parse_arithexpr(0);
-      break;
-
-    case 0: /* back-compat */
-    case XPI_OPERAND_TERM:
-      rhs = parse_termexpr(0);
+  switch(reg->hd.hooks->rhs_flags & 0x07) {
+    case 0:
       break;
 
     case XPI_OPERAND_TERM_LIST:
-      rhs = force_list_keeping_pushmark(parse_termexpr(0));
-      break;
-
     case XPI_OPERAND_LIST:
-      rhs = force_list_keeping_pushmark(parse_listexpr(0));
+      rhs = force_list_keeping_pushmark(rhs);
       break;
-
-    case XPI_OPERAND_CUSTOM:
-      rhs = (*reg->hd.hooks->parse_rhs)(aTHX_ reg->hd.data);
-      break;
-
-    default:
-      croak("hooks->rhs_flags did not provide a valid RHS type");
   }
 
-  return new_op(aTHX_ reg->hd, 0, lhs, rhs);
+  return new_op(aTHX_ reg->hd, 0, lhs, rhs, parsedata);
 }
 
 static STRLEN (*next_infix_plugin)(pTHX_ char *, STRLEN, struct Perl_custom_infix **);
@@ -380,7 +371,7 @@ OP *XSParseInfix_new_op(pTHX_ const struct XSParseInfixInfo *info, U32 flags, OP
     return new_op(aTHX_ (struct HooksAndData) {
         .hooks = info->hooks,
         .data  = info->hookdata,
-      }, flags, lhs, rhs);
+      }, flags, lhs, rhs, NULL);
 
   return newBINOP(info->opcode, flags, lhs, rhs);
 }
@@ -470,7 +461,7 @@ static OP *ckcall_wrapper_func_scalarscalar(pTHX_ OP *op, GV *namegv, SV *ckobj)
   if(!extract_wrapper2_args(aTHX_ op, &left, &right))
     return op;
 
-  return new_op(aTHX_ *hd, 0, left, right);
+  return new_op(aTHX_ *hd, 0, left, right, NULL);
 }
 
 static OP *ckcall_wrapper_func_listlist(pTHX_ OP *op, GV *namegv, SV *ckobj)
@@ -483,7 +474,8 @@ static OP *ckcall_wrapper_func_listlist(pTHX_ OP *op, GV *namegv, SV *ckobj)
 
   return new_op(aTHX_ *hd, 0,
       unwrap_list(left,  hd->hooks->lhs_flags & XPI_OPERAND_ONLY_LOOK),
-      unwrap_list(right, hd->hooks->rhs_flags & XPI_OPERAND_ONLY_LOOK));
+      unwrap_list(right, hd->hooks->rhs_flags & XPI_OPERAND_ONLY_LOOK),
+      NULL);
 }
 
 #define newSLUGOP(idx)  S_newSLUGOP(aTHX_ idx)
@@ -497,7 +489,9 @@ static OP *S_newSLUGOP(pTHX_ int idx)
 static void make_wrapper_func(pTHX_ const struct HooksAndData *hd)
 {
   SV *funcname = newSVpvn(hd->hooks->wrapper_func_name, strlen(hd->hooks->wrapper_func_name));
-  if(gv_fetchsv(funcname, 0, 0)) {
+
+  GV *gv;
+  if((gv = gv_fetchsv(funcname, 0, 0)) && GvCV(gv)) {
     /* The wrapper function already exists. We presume this is due to a duplicate
      * registration of identical hooks under a different name and just skip
      */
@@ -523,7 +517,7 @@ static void make_wrapper_func(pTHX_ const struct HooksAndData *hd)
 
       /* Body of the function is just  $_[0] OP $_[1] */
       body = op_append_list(OP_LINESEQ, body,
-          new_op(aTHX_ *hd, 0, newSLUGOP(0), newSLUGOP(1)));
+          new_op(aTHX_ *hd, 0, newSLUGOP(0), newSLUGOP(1), NULL));
 
       ckcall = &ckcall_wrapper_func_scalarscalar;
       break;
@@ -539,7 +533,8 @@ static void make_wrapper_func(pTHX_ const struct HooksAndData *hd)
       body = op_append_list(OP_LINESEQ, body,
           new_op(aTHX_ *hd, 0,
             newOP(OP_SHIFT, 0),
-            force_list_keeping_pushmark(newUNOP(OP_RV2AV, OPf_WANT_LIST, newGVOP(OP_GV, 0, PL_defgv)))));
+            force_list_keeping_pushmark(newUNOP(OP_RV2AV, OPf_WANT_LIST, newGVOP(OP_GV, 0, PL_defgv))),
+            NULL));
 
       /* no ckcall */
       break;
@@ -555,7 +550,8 @@ static void make_wrapper_func(pTHX_ const struct HooksAndData *hd)
       body = op_append_list(OP_LINESEQ, body,
           new_op(aTHX_ *hd, 0,
             force_list_keeping_pushmark(newUNOP(OP_RV2AV, 0, newSLUGOP(0))),
-            force_list_keeping_pushmark(newUNOP(OP_RV2AV, 0, newSLUGOP(1)))));
+            force_list_keeping_pushmark(newUNOP(OP_RV2AV, 0, newSLUGOP(1))),
+            NULL));
 
       ckcall = &ckcall_wrapper_func_listlist;
       break;
@@ -665,6 +661,8 @@ static void reg_builtin(pTHX_ const char *opname, enum XSParseInfixClassificatio
 void XSParseInfix_register(pTHX_ const char *opname, const struct XSParseInfixHooks *hooks, void *hookdata)
 {
   switch(hooks->flags) {
+    case (1<<15):
+      /* undocumented internal flag to indicate v1-compatible ->new_op hook function */
     case 0:
       break;
     default:
@@ -673,10 +671,8 @@ void XSParseInfix_register(pTHX_ const char *opname, const struct XSParseInfixHo
 
   switch(hooks->lhs_flags & ~(XPI_OPERAND_ONLY_LOOK)) {
     case 0:
-      //warn("No .lhs_flags specified for XSParseInfixHooks; defaulting to XPI_OPERAND_TERM");
-      /* FALLTHROUGH */
-    case XPI_OPERAND_TERM:
     case XPI_OPERAND_TERM_LIST:
+    case XPI_OPERAND_LIST:
       break;
     default:
       croak("Unrecognised XSParseInfixHooks.lhs_flags value 0x%X", hooks->lhs_flags);
@@ -684,23 +680,57 @@ void XSParseInfix_register(pTHX_ const char *opname, const struct XSParseInfixHo
 
   switch(hooks->rhs_flags & ~(XPI_OPERAND_ONLY_LOOK)) {
     case 0:
-      //warn("No .rhs_flags specified for XSParseInfixHooks; defaulting to XPI_OPERAND_TERM");
-      /* FALLTHROUGH */
-    case XPI_OPERAND_ARITH:
-    case XPI_OPERAND_TERM:
     case XPI_OPERAND_TERM_LIST:
     case XPI_OPERAND_LIST:
-    case XPI_OPERAND_CUSTOM:
       break;
     default:
       croak("Unrecognised XSParseInfixHooks.rhs_flags value 0x%X", hooks->rhs_flags);
+
+    case XPI_OPERAND_CUSTOM:
+      croak("TODO: Currently XPI_OPERAND_CUSTOM is not supported");
   }
 
+#ifdef HAVE_PL_INFIX_PLUGIN
+  enum Perl_custom_infix_precedence prec = 0;
+
+  switch(hooks->cls) {
+    case 0:
+      warn("Unspecified operator classification for %s; treating it as RELATION for precedence", opname);
+    case XPI_CLS_RELATION:
+    case XPI_CLS_EQUALITY:
+    case XPI_CLS_MATCH_MISC:
+      prec = INFIX_PREC_REL;
+      break;
+
+    case XPI_CLS_ADD_MISC:
+      prec = INFIX_PREC_ADD;
+      break;
+
+    case XPI_CLS_MUL_MISC:
+      prec = INFIX_PREC_MUL;
+      break;
+
+    case XPI_CLS_POW_MISC:
+      prec = INFIX_PREC_POW;
+      break;
+
+    /* TODO: Add classifications for the HIGH and LOW also? */
+
+    default:
+      croak("TODO: need to write code for hooks->cls == %d\n", hooks->cls);
+  }
+#endif
+
   struct Registration *reg;
   Newx(reg, 1, struct Registration);
 
 #ifdef HAVE_PL_INFIX_PLUGIN
-  reg->def.parse = &parse;
+  reg->def.prec  = prec;
+  if(hooks->parse)
+    reg->def.parse = &parse;
+  else
+    reg->def.parse = NULL;
+  reg->def.build_op = &build_op;
 #endif
 
   reg->info.opname = savepv(opname);
